@@ -77,6 +77,7 @@ struct a2dp_sep {
 };
 
 struct a2dp_setup_cb {
+	struct a2dp_setup *setup;
 	a2dp_select_cb_t select_cb;
 	a2dp_config_cb_t config_cb;
 	a2dp_stream_cb_t resume_cb;
@@ -93,7 +94,7 @@ struct a2dp_setup {
 	struct avdtp_stream *stream;
 	struct avdtp_error *err;
 	avdtp_set_configuration_cb setconf_cb;
-	GSList *client_caps;
+	GSList *caps;
 	gboolean reconfigure;
 	gboolean start;
 	GSList *cb;
@@ -124,27 +125,6 @@ static struct a2dp_setup *setup_ref(struct a2dp_setup *setup)
 	return setup;
 }
 
-static void setup_free(struct a2dp_setup *s)
-{
-	DBG("%p", s);
-	setups = g_slist_remove(setups, s);
-	if (s->session)
-		avdtp_unref(s->session);
-	g_slist_foreach(s->cb, (GFunc) g_free, NULL);
-	g_slist_free(s->cb);
-	g_free(s);
-}
-
-static void setup_unref(struct a2dp_setup *setup)
-{
-	setup->ref--;
-
-	DBG("%p: ref=%d", setup, setup->ref);
-
-	if (setup->ref <= 0)
-		setup_free(setup);
-}
-
 static struct audio_device *a2dp_get_dev(struct avdtp *session)
 {
 	bdaddr_t src, dst;
@@ -154,26 +134,90 @@ static struct audio_device *a2dp_get_dev(struct avdtp *session)
 	return manager_find_device(NULL, &src, &dst, NULL, FALSE);
 }
 
+static struct a2dp_setup *setup_new(struct avdtp *session)
+{
+	struct audio_device *dev;
+	struct a2dp_setup *setup;
+
+	dev = a2dp_get_dev(session);
+	if (!dev) {
+		error("Unable to create setup");
+		return NULL;
+	}
+
+	setup = g_new0(struct a2dp_setup, 1);
+	setup->session = avdtp_ref(session);
+	setup->dev = a2dp_get_dev(session);
+	setups = g_slist_append(setups, setup);
+
+	return setup;
+}
+
+static void setup_free(struct a2dp_setup *s)
+{
+	DBG("%p", s);
+
+	setups = g_slist_remove(setups, s);
+	if (s->session)
+		avdtp_unref(s->session);
+	g_slist_foreach(s->cb, (GFunc) g_free, NULL);
+	g_slist_free(s->cb);
+	g_slist_foreach(s->caps, (GFunc) g_free, NULL);
+	g_slist_free(s->caps);
+	g_free(s);
+}
+
+static void setup_unref(struct a2dp_setup *setup)
+{
+	setup->ref--;
+
+	DBG("%p: ref=%d", setup, setup->ref);
+
+	if (setup->ref > 0)
+		return;
+
+	setup_free(setup);
+}
+
+static struct a2dp_setup_cb *setup_cb_new(struct a2dp_setup *setup)
+{
+	struct a2dp_setup_cb *cb;
+
+	cb = g_new0(struct a2dp_setup_cb, 1);
+	cb->setup = setup;
+	cb->id = ++cb_id;
+
+	setup->cb = g_slist_append(setup->cb, cb);
+	return cb;
+}
+
+static void setup_cb_free(struct a2dp_setup_cb *cb)
+{
+	struct a2dp_setup *setup = cb->setup;
+
+	setup->cb = g_slist_remove(setup->cb, cb);
+	setup_unref(cb->setup);
+	g_free(cb);
+}
+
 static gboolean finalize_config(struct a2dp_setup *s)
 {
 	GSList *l;
 	struct avdtp_stream *stream = s->err ? NULL : s->stream;
 
-	setup_ref(s);
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
+		l = l->next;
+
 		if (!cb->config_cb)
 			continue;
 
 		cb->config_cb(s->session, s->sep, stream, s->err,
 							cb->user_data);
-		cb->config_cb = NULL;
-		setup_unref(s);
+		setup_cb_free(cb);
 	}
 
-	setup_unref(s);
-
 	return FALSE;
 }
 
@@ -191,19 +235,17 @@ static gboolean finalize_resume(struct a2dp_setup *s)
 {
 	GSList *l;
 
-	setup_ref(s);
-
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb && cb->resume_cb) {
-			cb->resume_cb(s->session, s->err, cb->user_data);
-			cb->resume_cb = NULL;
-			setup_unref(s);
-		}
-	}
+		l = l->next;
 
-	setup_unref(s);
+		if (!cb->resume_cb)
+			continue;
+
+		cb->resume_cb(s->session, s->err, cb->user_data);
+		setup_cb_free(cb);
+	}
 
 	return FALSE;
 }
@@ -212,18 +254,18 @@ static gboolean finalize_suspend(struct a2dp_setup *s)
 {
 	GSList *l;
 
-	setup_ref(s);
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->suspend_cb) {
-			cb->suspend_cb(s->session, s->err, cb->user_data);
-			cb->suspend_cb = NULL;
-			setup_unref(s);
-		}
+		l = l->next;
+
+		if (!cb->suspend_cb)
+			continue;
+
+		cb->suspend_cb(s->session, s->err, cb->user_data);
+		setup_cb_free(cb);
 	}
 
-	setup_unref(s);
 	return FALSE;
 }
 
@@ -237,22 +279,22 @@ static gboolean finalize_suspend_errno(struct a2dp_setup *s, int err)
 	return finalize_suspend(s);
 }
 
-static gboolean finalize_select(struct a2dp_setup *s, GSList *caps)
+static gboolean finalize_select(struct a2dp_setup *s)
 {
 	GSList *l;
 
-	setup_ref(s);
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->select_cb) {
-			cb->select_cb(s->session, s->sep, caps, cb->user_data);
-			cb->select_cb = NULL;
-			setup_unref(s);
-		}
+		l = l->next;
+
+		if (!cb->select_cb)
+			continue;
+
+		cb->select_cb(s->session, s->sep, s->caps, cb->user_data);
+		setup_cb_free(cb);
 	}
 
-	setup_unref(s);
 	return FALSE;
 }
 
@@ -270,6 +312,20 @@ static struct a2dp_setup *find_setup_by_session(struct avdtp *session)
 	return NULL;
 }
 
+static struct a2dp_setup *a2dp_setup_get(struct avdtp *session)
+{
+	struct a2dp_setup *setup;
+
+	setup = find_setup_by_session(session);
+	if (!setup) {
+		setup = setup_new(session);
+		if (!setup)
+			return NULL;
+	}
+
+	return setup_ref(setup);
+}
+
 static struct a2dp_setup *find_setup_by_dev(struct audio_device *dev)
 {
 	GSList *l;
@@ -343,6 +399,8 @@ done:
 	if (err)
 		g_free(err);
 
+	setup_unref(setup);
+
 	return FALSE;
 }
 
@@ -354,7 +412,6 @@ static gboolean sbc_setconf_ind(struct avdtp *session,
 					void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct audio_device *dev;
 	struct a2dp_setup *setup;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
@@ -362,19 +419,10 @@ static gboolean sbc_setconf_ind(struct avdtp *session,
 	else
 		DBG("Source %p: Set_Configuration_Ind", sep);
 
-	dev = a2dp_get_dev(session);
-	if (!dev)
+	setup = a2dp_setup_get(session);
+	if (!setup)
 		return FALSE;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = dev;
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
 	a2dp_sep->stream = stream;
 	setup->sep = a2dp_sep;
 	setup->stream = stream;
@@ -492,7 +540,6 @@ static gboolean mpeg_setconf_ind(struct avdtp *session,
 					void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct audio_device *dev;
 	struct a2dp_setup *setup;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
@@ -500,19 +547,10 @@ static gboolean mpeg_setconf_ind(struct avdtp *session,
 	else
 		DBG("Source %p: Set_Configuration_Ind", sep);
 
-	dev = a2dp_get_dev(session);
-	if (!dev)
+	setup = a2dp_setup_get(session);
+	if (!setup)
 		return FALSE;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = dev;
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
 	a2dp_sep->stream = stream;
 	setup->sep = a2dp_sep;
 	setup->stream = stream;
@@ -614,7 +652,6 @@ static gboolean endpoint_setconf_ind(struct avdtp *session,
 						void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct audio_device *dev;
 	struct a2dp_setup *setup;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
@@ -622,19 +659,10 @@ static gboolean endpoint_setconf_ind(struct avdtp *session,
 	else
 		DBG("Source %p: Set_Configuration_Ind", sep);
 
-	dev = a2dp_get_dev(session);
-	if (!dev)
+	setup = a2dp_setup_get(session);
+	if (!session)
 		return FALSE;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = dev;
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
 	a2dp_sep->stream = stream;
 	setup->sep = a2dp_sep;
 	setup->stream = stream;
@@ -665,9 +693,10 @@ static gboolean endpoint_setconf_ind(struct avdtp *session,
 			goto done;
 		}
 
-		ret = media_endpoint_set_configuration(a2dp_sep->endpoint, dev,
-					codec->data, cap->length - sizeof(*codec),
-					endpoint_setconf_cb, setup);
+		ret = media_endpoint_set_configuration(a2dp_sep->endpoint,
+						setup->dev, codec->data,
+						cap->length - sizeof(*codec),
+						endpoint_setconf_cb, setup);
 		if (ret)
 			return TRUE;
 
@@ -1033,7 +1062,7 @@ static gboolean a2dp_reconfigure(gpointer data)
 
 	posix_err = avdtp_set_configuration(setup->session, setup->rsep,
 						sep->lsep,
-						setup->client_caps,
+						setup->caps,
 						&setup->stream);
 	if (posix_err < 0) {
 		error("avdtp_set_configuration: %s", strerror(-posix_err));
@@ -1070,6 +1099,9 @@ static void close_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
 		return;
 	}
 
+	if (!setup->rsep)
+		setup->rsep = avdtp_stream_get_remote_sep(stream);
+
 	if (setup->reconfigure)
 		g_timeout_add(RECONFIGURE_TIMEOUT, a2dp_reconfigure, setup);
 }
@@ -1784,7 +1816,6 @@ static void select_cb(struct media_endpoint *endpoint, void *ret, int size,
 	struct a2dp_setup *setup = user_data;
 	struct avdtp_service_capability *media_transport, *media_codec;
 	struct avdtp_media_codec_capability *cap;
-	GSList *caps = NULL;
 
 	if (size < 0) {
 		DBG("Endpoint replied an invalid configuration");
@@ -1794,7 +1825,7 @@ static void select_cb(struct media_endpoint *endpoint, void *ret, int size,
 	media_transport = avdtp_service_cap_new(AVDTP_MEDIA_TRANSPORT,
 						NULL, 0);
 
-	caps = g_slist_append(caps, media_transport);
+	setup->caps = g_slist_append(setup->caps, media_transport);
 
 	cap = g_malloc0(sizeof(*cap) + size);
 	cap->media_type = AVDTP_MEDIA_TYPE_AUDIO;
@@ -1804,17 +1835,18 @@ static void select_cb(struct media_endpoint *endpoint, void *ret, int size,
 	media_codec = avdtp_service_cap_new(AVDTP_MEDIA_CODEC, cap,
 						sizeof(*cap) + size);
 
-	caps = g_slist_append(caps, media_codec);
+	setup->caps = g_slist_append(setup->caps, media_codec);
+	g_free(cap);
 
 done:
-	finalize_select(setup, caps);
+	finalize_select(setup);
 }
 
 static gboolean auto_select(gpointer data)
 {
 	struct a2dp_setup *setup = data;
 
-	finalize_select(setup, setup->client_caps);
+	finalize_select(setup);
 
 	return FALSE;
 }
@@ -1886,21 +1918,14 @@ unsigned int a2dp_select_capabilities(struct avdtp *session,
 		return 0;
 	}
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->select_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->rsep = avdtp_find_remote_sep(session, sep->lsep);
 
@@ -1913,7 +1938,7 @@ unsigned int a2dp_select_capabilities(struct avdtp *session,
 	endpoint in the configuration file */
 	if (sep->endpoint == NULL) {
 		if (!select_capabilities(session, setup->rsep,
-					&setup->client_caps)) {
+					&setup->caps)) {
 			error("Unable to auto select remote SEP capabilities");
 			goto fail;
 		}
@@ -1933,8 +1958,7 @@ unsigned int a2dp_select_capabilities(struct avdtp *session,
 		return cb_data->id;
 
 fail:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 
 }
@@ -1977,24 +2001,23 @@ unsigned int a2dp_config(struct avdtp *session, struct a2dp_sep *sep,
 
 	DBG("a2dp_config: selected SEP %p", sep->lsep);
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->config_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->stream = sep->stream;
-	setup->client_caps = caps;
+
+	/* Copy given caps if they are different than current caps */
+	if (setup->caps != caps) {
+		g_slist_foreach(setup->caps, (GFunc) g_free, NULL);
+		g_slist_free(setup->caps);
+		setup->caps = g_slist_copy(caps);
+	}
 
 	switch (avdtp_sep_get_state(sep->lsep)) {
 	case AVDTP_STATE_IDLE:
@@ -2059,8 +2082,7 @@ unsigned int a2dp_config(struct avdtp *session, struct a2dp_sep *sep,
 	return cb_data->id;
 
 failed:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 }
 
@@ -2070,21 +2092,14 @@ unsigned int a2dp_resume(struct avdtp *session, struct a2dp_sep *sep,
 	struct a2dp_setup_cb *cb_data;
 	struct a2dp_setup *setup;
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->resume_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->stream = sep->stream;
 
@@ -2118,8 +2133,7 @@ unsigned int a2dp_resume(struct avdtp *session, struct a2dp_sep *sep,
 	return cb_data->id;
 
 failed:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 }
 
@@ -2129,21 +2143,14 @@ unsigned int a2dp_suspend(struct avdtp *session, struct a2dp_sep *sep,
 	struct a2dp_setup_cb *cb_data;
 	struct a2dp_setup *setup;
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->suspend_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->stream = sep->stream;
 
@@ -2170,44 +2177,39 @@ unsigned int a2dp_suspend(struct avdtp *session, struct a2dp_sep *sep,
 	return cb_data->id;
 
 failed:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 }
 
 gboolean a2dp_cancel(struct audio_device *dev, unsigned int id)
 {
-	struct a2dp_setup_cb *cb_data;
 	struct a2dp_setup *setup;
 	GSList *l;
 
-	DBG("a2dp_cancel()");
-
 	setup = find_setup_by_dev(dev);
 	if (!setup)
 		return FALSE;
 
-	for (cb_data = NULL, l = setup->cb; l != NULL; l = g_slist_next(l)) {
+	for (l = setup->cb; l != NULL; l = g_slist_next(l)) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->id == id) {
-			cb_data = cb;
-			break;
+		if (cb->id != id)
+			continue;
+
+		setup_ref(setup);
+		setup_cb_free(cb);
+
+		if (!setup->cb) {
+			DBG("aborting setup %p", setup);
+			avdtp_abort(setup->session, setup->stream);
+			return TRUE;
 		}
+
+		setup_unref(setup);
+		return TRUE;
 	}
 
-	if (!cb_data)
-		error("a2dp_cancel: no matching callback with id %u", id);
-
-	setup->cb = g_slist_remove(setup->cb, cb_data);
-	g_free(cb_data);
-
-	if (setup->cb)
-		return TRUE;
-
-	avdtp_abort(setup->session, setup->stream);
-
-	return TRUE;
+	return FALSE;
 }
 
 gboolean a2dp_sep_lock(struct a2dp_sep *sep, struct avdtp *session)
